% DUALNUMBER  Class that implements dual numbers required for 
% automatic differentiation. You cand construct dual number using
% three different ways:
%
%   obj = DualNumber(a, b) % Construct a + bEPS value
%   obj = DualNumber.CreateXValue(a) % Constructs a + 1EPS value
%   obj = DualNumber.CreteConst(a) % Construct a + 0EPS value
%
%   a and b are double scalars or matrices
%
% It supports operators overloading (+, -, *, ./, .^) and allows to
% write expressions in a natural algebraic way, e.g.:
%   (2.*x+3).^(3.*x-1)
classdef DualNumber
    properties
        a
        b
    end
    
    methods
        function obj = DualNumber(a, b)
            if ~isa(a, 'double') || ~isreal(a)
                error('a must be a real double matrix');
            end
            if ~isa(b, 'double') || ~isreal(b)
                error('b must be a real double matrix');
            end
            if any(size(a) ~= size(b))
                error('a and b sizes are not consistent');
            end
            obj.a = a;
            obj.b = b;
        end
        
        function r = plus(o1, o2)
            [o1, o2] = prepareBinArgs(o1, o2);
            r = DualNumber(o1.a + o2.a, o1.b + o2.b);
        end
        
        function r = minus(o1, o2)
            [o1, o2] = prepareBinArgs(o1, o2);
            r = DualNumber(o1.a - o2.a, o1.b - o2.b);
        end
        
        function r = uminus(o1)
            r = DualNumber(-o1.a, -o1.b);
        end
        
        function r = uplus(o1)
            r = DualNumber(+o1.a, +o1.b);
        end
        
        function r = times(o1, o2)
            [o1, o2] = prepareBinArgs(o1, o2);           
            r = DualNumber(o1.a.*o2.a, o1.b.*o2.a + o1.a.*o2.b);
        end
        
        function r = mtimes(o1, o2)
            [o1, o2] = prepareBinArgs(o1, o2);
            r = DualNumber(o1.a*o2.a, o1.b*o2.a + o1.a*o2.b);
        end
        
        function r = power(o1, o2)
            [o1, o2] = prepareBinArgs(o1, o2);
            r = DualNumber(o1.a.^o2.a, o1.a.^o2.a.*((o1.b.*o2.a)./o1.a + o2.b.*log(o1.a)));
        end
        
        function r = rdivide(o1, o2)
            [o1, o2] = prepareBinArgs(o1, o2);            
            r = DualNumber(o1.a./o2.a, (o1.b.*o2.a - o1.a.*o2.b)./(o2.a).^2);
        end
        
        function r = transpose(o1)
            r = DualNumber(o1.a.', o1.b.');
        end
        
        function r = ctranspose(o1)
            r = DualNumber(o1.a', o1.b');
        end
        
        function r = abs(o1)
        % Calculate absolute value of dual numbers
            r = builtin('abs', o1.a);
        end
        
        function r = sum(o1, dim)
            mata = builtin('sum', o1.a, dim);
            matb = builtin('sum', o1.b, dim);
            r = DualNumber(mata, matb);
        end
        
        function r = log(o1)
            r = DualNumber(builtin('log', o1.a), o1.b./o1.a);
        end
        
        function r = real(o1)
            r = o1.a;
        end
        
        function r = imag(o1)
            r = o1.b;
        end
                
        function r = horzcat(o1,varargin)
            [vara, varb] = prepareCat(o1, 1, varargin);
            amat = horzcat(o1.a, vara{:});
            bmat = horzcat(o1.b, varb{:});
            r = DualNumber(amat, bmat);
        end
        
        function r = vertcat(o1,varargin)
            [vara, varb] = prepareCat(o1, 2, varargin);
            amat = vertcat(o1.a, vara{:});
            bmat = vertcat(o1.b, varb{:});
            r = DualNumber(amat, bmat);
        end        
        
        function r = unwrapDouble(obj, num)
            if ~isa(num, 'double')
                error('num argument must be a double matrix');
            end
            r = DualNumber.CreateConst(num);
        end
        
        function [o1, o2] = prepareBinArgs(o1, o2)
            if isa(o1, 'double')
                o1 = o2.unwrapDouble(o1);
            elseif isa(o2, 'double')
                o2 = o1.unwrapDouble(o2);
            end            
        end
        
        function [vara, varb] = prepareCat(o1, dim, vars)
            vara = cell(size(vars));
            varb = cell(size(vars));
            for i = 1:numel(vars)
                var = vars{i};
                if ~isa(var, 'DualNumber')
                    error('All variables must be DualNumber');
                end
                if size(var.a, dim) ~= size(o1.a, dim)
                    error('Dimensions of matrices being concatenated are not consistent');
                end
                vara{i} = var.a;
                varb{i} = var.b;
            end
        end
        
        function disp(obj)
            [m, n] = size(obj.a);            
            divisor = max(max(max(abs(obj.a))), max(max(abs(obj.b))));
            if divisor == 0
                divisor_pow = 0;
            else
                divisor_pow = floor(log10(divisor));
            end
            fprintf('%d x %d DualNumber matrix (@ means epsilon)\n', m, n);
            if divisor_pow ~= 0
                fprintf('  %.2e *\n', 10.^divisor_pow);
            end
            for i=1:m
                fprintf('   ');
                for j=1:n
                    ai = obj.a(i,j)./10.^divisor_pow;
                    bi = obj.b(i,j)./10.^divisor_pow;
                    if bi >= 0
                        fprintf('%7.4f+%6.4f@ ', ai, bi);
                    else
                        fprintf('%7.4f-%6.4f@ ', ai, -bi);
                    end
                end
                fprintf('\n');
            end
            fprintf('\n');
        end
    end
    
    methods (Static)
        function obj = CreateConst(a)
            obj = DualNumber(a, zeros(size(a)));
        end
        
        function obj = CreateXValue(a)
            obj = DualNumber(a, ones(size(a)));
        end
    end
end